import datetime
from io import BytesIO
import os
import shutil

import numpy as np
import pytest

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.testing import _has_tex_package, _check_for_pgf
from matplotlib.testing.compare import compare_images, ImageComparisonFailure
from matplotlib.backends.backend_pgf import PdfPages, common_texification
from matplotlib.testing.decorators import (_image_directories,
                                           check_figures_equal,
                                           image_comparison)

baseline_dir, result_dir = _image_directories(lambda: 'dummy func')

needs_xelatex = pytest.mark.skipif(not _check_for_pgf('xelatex'),
                                   reason='xelatex + pgf is required')
needs_pdflatex = pytest.mark.skipif(not _check_for_pgf('pdflatex'),
                                    reason='pdflatex + pgf is required')
needs_lualatex = pytest.mark.skipif(not _check_for_pgf('lualatex'),
                                    reason='lualatex + pgf is required')
needs_ghostscript = pytest.mark.skipif(
    "eps" not in mpl.testing.compare.converter,
    reason="This test needs a ghostscript installation")


def compare_figure(fname, savefig_kwargs={}, tol=0):
    actual = os.path.join(result_dir, fname)
    plt.savefig(actual, **savefig_kwargs)

    expected = os.path.join(result_dir, "expected_%s" % fname)
    shutil.copyfile(os.path.join(baseline_dir, fname), expected)
    err = compare_images(expected, actual, tol=tol)
    if err:
        raise ImageComparisonFailure(err)


def create_figure():
    plt.figure()
    x = np.linspace(0, 1, 15)

    # line plot
    plt.plot(x, x ** 2, "b-")

    # marker
    plt.plot(x, 1 - x**2, "g>")

    # filled paths and patterns
    plt.fill_between([0., .4], [.4, 0.], hatch='//', facecolor="lightgray",
                     edgecolor="red")
    plt.fill([3, 3, .8, .8, 3], [2, -2, -2, 0, 2], "b")

    # text and typesetting
    plt.plot([0.9], [0.5], "ro", markersize=3)
    plt.text(0.9, 0.5, 'unicode (ü, °, µ) and math ($\\mu_i = x_i^2$)',
             ha='right', fontsize=20)
    plt.ylabel('sans-serif, blue, $\\frac{\\sqrt{x}}{y^2}$..',
               family='sans-serif', color='blue')

    plt.xlim(0, 1)
    plt.ylim(0, 1)


@pytest.mark.parametrize('plain_text, escaped_text', [
    (r'quad_sum: $\sum x_i^2$', r'quad\_sum: \(\displaystyle \sum x_i^2\)'),
    (r'no \$splits \$ here', r'no \$splits \$ here'),
    ('with_underscores', r'with\_underscores'),
    ('% not a comment', r'\% not a comment'),
    ('^not', r'\^not'),
])
def test_common_texification(plain_text, escaped_text):
    assert common_texification(plain_text) == escaped_text


# test compiling a figure to pdf with xelatex
@needs_xelatex
@pytest.mark.backend('pgf')
@image_comparison(['pgf_xelatex.pdf'], style='default')
def test_xelatex():
    rc_xelatex = {'font.family': 'serif',
                  'pgf.rcfonts': False}
    mpl.rcParams.update(rc_xelatex)
    create_figure()


# test compiling a figure to pdf with pdflatex
@needs_pdflatex
@pytest.mark.skipif(not _has_tex_package('ucs'), reason='needs ucs.sty')
@pytest.mark.backend('pgf')
@image_comparison(['pgf_pdflatex.pdf'], style='default')
def test_pdflatex():
    if os.environ.get('APPVEYOR'):
        pytest.xfail("pdflatex test does not work on appveyor due to missing "
                     "LaTeX fonts")

    rc_pdflatex = {'font.family': 'serif',
                   'pgf.rcfonts': False,
                   'pgf.texsystem': 'pdflatex',
                   'pgf.preamble': ('\\usepackage[utf8x]{inputenc}'
                                    '\\usepackage[T1]{fontenc}')}
    mpl.rcParams.update(rc_pdflatex)
    create_figure()


# test updating the rc parameters for each figure
@needs_xelatex
@needs_pdflatex
@pytest.mark.style('default')
@pytest.mark.backend('pgf')
def test_rcupdate():
    rc_sets = [{'font.family': 'sans-serif',
                'font.size': 30,
                'figure.subplot.left': .2,
                'lines.markersize': 10,
                'pgf.rcfonts': False,
                'pgf.texsystem': 'xelatex'},
               {'font.family': 'monospace',
                'font.size': 10,
                'figure.subplot.left': .1,
                'lines.markersize': 20,
                'pgf.rcfonts': False,
                'pgf.texsystem': 'pdflatex',
                'pgf.preamble': ('\\usepackage[utf8x]{inputenc}'
                                 '\\usepackage[T1]{fontenc}'
                                 '\\usepackage{sfmath}')}]
    tol = [6, 0]
    for i, rc_set in enumerate(rc_sets):
        with mpl.rc_context(rc_set):
            for substring, pkg in [('sfmath', 'sfmath'), ('utf8x', 'ucs')]:
                if (substring in mpl.rcParams['pgf.preamble']
                        and not _has_tex_package(pkg)):
                    pytest.skip(f'needs {pkg}.sty')
            create_figure()
            compare_figure('pgf_rcupdate%d.pdf' % (i + 1), tol=tol[i])


# test backend-side clipping, since large numbers are not supported by TeX
@needs_xelatex
@pytest.mark.style('default')
@pytest.mark.backend('pgf')
def test_pathclip():
    mpl.rcParams.update({'font.family': 'serif', 'pgf.rcfonts': False})
    plt.plot([0., 1e100], [0., 1e100])
    plt.xlim(0, 1)
    plt.ylim(0, 1)
    plt.savefig(BytesIO(), format="pdf")  # No image comparison.


# test mixed mode rendering
@needs_xelatex
@pytest.mark.backend('pgf')
@image_comparison(['pgf_mixedmode.pdf'], style='default')
def test_mixedmode():
    mpl.rcParams.update({'font.family': 'serif', 'pgf.rcfonts': False})
    Y, X = np.ogrid[-1:1:40j, -1:1:40j]
    plt.pcolor(X**2 + Y**2).set_rasterized(True)


# test bbox_inches clipping
@needs_xelatex
@pytest.mark.style('default')
@pytest.mark.backend('pgf')
def test_bbox_inches():
    mpl.rcParams.update({'font.family': 'serif', 'pgf.rcfonts': False})
    fig, (ax1, ax2) = plt.subplots(1, 2)
    ax1.plot(range(5))
    ax2.plot(range(5))
    plt.tight_layout()
    bbox = ax1.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    compare_figure('pgf_bbox_inches.pdf', savefig_kwargs={'bbox_inches': bbox},
                   tol=0)


@pytest.mark.style('default')
@pytest.mark.backend('pgf')
@pytest.mark.parametrize('system', [
    pytest.param('lualatex', marks=[needs_lualatex]),
    pytest.param('pdflatex', marks=[needs_pdflatex]),
    pytest.param('xelatex', marks=[needs_xelatex]),
])
def test_pdf_pages(system):
    rc_pdflatex = {
        'font.family': 'serif',
        'pgf.rcfonts': False,
        'pgf.texsystem': system,
    }
    mpl.rcParams.update(rc_pdflatex)

    fig1, ax1 = plt.subplots()
    ax1.plot(range(5))
    fig1.tight_layout()

    fig2, ax2 = plt.subplots(figsize=(3, 2))
    ax2.plot(range(5))
    fig2.tight_layout()

    path = os.path.join(result_dir, f'pdfpages_{system}.pdf')
    md = {
        'Author': 'me',
        'Title': 'Multipage PDF with pgf',
        'Subject': 'Test page',
        'Keywords': 'test,pdf,multipage',
        'ModDate': datetime.datetime(
            1968, 8, 1, tzinfo=datetime.timezone(datetime.timedelta(0))),
        'Trapped': 'Unknown'
    }

    with PdfPages(path, metadata=md) as pdf:
        pdf.savefig(fig1)
        pdf.savefig(fig2)
        pdf.savefig(fig1)

        assert pdf.get_pagecount() == 3


@pytest.mark.style('default')
@pytest.mark.backend('pgf')
@pytest.mark.parametrize('system', [
    pytest.param('lualatex', marks=[needs_lualatex]),
    pytest.param('pdflatex', marks=[needs_pdflatex]),
    pytest.param('xelatex', marks=[needs_xelatex]),
])
def test_pdf_pages_metadata_check(monkeypatch, system):
    # Basically the same as test_pdf_pages, but we keep it separate to leave
    # pikepdf as an optional dependency.
    pikepdf = pytest.importorskip('pikepdf')
    monkeypatch.setenv('SOURCE_DATE_EPOCH', '0')

    mpl.rcParams.update({'pgf.texsystem': system})

    fig, ax = plt.subplots()
    ax.plot(range(5))

    md = {
        'Author': 'me',
        'Title': 'Multipage PDF with pgf',
        'Subject': 'Test page',
        'Keywords': 'test,pdf,multipage',
        'ModDate': datetime.datetime(
            1968, 8, 1, tzinfo=datetime.timezone(datetime.timedelta(0))),
        'Trapped': 'True'
    }
    path = os.path.join(result_dir, f'pdfpages_meta_check_{system}.pdf')
    with PdfPages(path, metadata=md) as pdf:
        pdf.savefig(fig)

    with pikepdf.Pdf.open(path) as pdf:
        info = {k: str(v) for k, v in pdf.docinfo.items()}

    # Not set by us, so don't bother checking.
    if '/PTEX.FullBanner' in info:
        del info['/PTEX.FullBanner']
    if '/PTEX.Fullbanner' in info:
        del info['/PTEX.Fullbanner']

    assert info == {
        '/Author': 'me',
        '/CreationDate': 'D:19700101000000Z',
        '/Creator': f'Matplotlib v{mpl.__version__}, https://matplotlib.org',
        '/Keywords': 'test,pdf,multipage',
        '/ModDate': 'D:19680801000000Z',
        '/Producer': f'Matplotlib pgf backend v{mpl.__version__}',
        '/Subject': 'Test page',
        '/Title': 'Multipage PDF with pgf',
        '/Trapped': '/True',
    }


@needs_xelatex
def test_tex_restart_after_error():
    fig = plt.figure()
    fig.suptitle(r"\oops")
    with pytest.raises(ValueError):
        fig.savefig(BytesIO(), format="pgf")

    fig = plt.figure()  # start from scratch
    fig.suptitle(r"this is ok")
    fig.savefig(BytesIO(), format="pgf")


@needs_xelatex
def test_bbox_inches_tight():
    fig, ax = plt.subplots()
    ax.imshow([[0, 1], [2, 3]])
    fig.savefig(BytesIO(), format="pdf", backend="pgf", bbox_inches="tight")


@needs_xelatex
@needs_ghostscript
def test_png():
    # Just a smoketest.
    fig, ax = plt.subplots()
    fig.savefig(BytesIO(), format="png", backend="pgf")


@needs_xelatex
def test_unknown_font(caplog):
    with caplog.at_level("WARNING"):
        mpl.rcParams["font.family"] = "this-font-does-not-exist"
        plt.figtext(.5, .5, "hello, world")
        plt.savefig(BytesIO(), format="pgf")
    assert "Ignoring unknown font: this-font-does-not-exist" in [
        r.getMessage() for r in caplog.records]


@check_figures_equal(extensions=["pdf"])
@pytest.mark.parametrize("texsystem", ("pdflatex", "xelatex", "lualatex"))
@pytest.mark.backend("pgf")
def test_minus_signs_with_tex(fig_test, fig_ref, texsystem):
    if not _check_for_pgf(texsystem):
        pytest.skip(texsystem + ' + pgf is required')
    mpl.rcParams["pgf.texsystem"] = texsystem
    fig_test.text(.5, .5, "$-1$")
    fig_ref.text(.5, .5, "$\N{MINUS SIGN}1$")



import logging
from typing import Callable, Any

from rx import empty
from rx.core import Observable
from rx.internal import noop
from rx.internal.utils import add_ref
from rx.disposable import SingleAssignmentDisposable, SerialDisposable, CompositeDisposable, RefCountDisposable
from rx.subject import Subject
from rx import operators as ops

log = logging.getLogger("Rx")


def _window_toggle(openings: Observable,
                   closing_mapper: Callable[[Any], Observable]
                   ) -> Callable[[Observable], Observable]:
    """Projects each element of an observable sequence into zero or
    more windows.

    Args:
        source: Source observable to project into windows.

    Returns:
        An observable sequence of windows.
    """
    def window_toggle(source: Observable) -> Observable:

        def mapper(args):
            _, window = args
            return window

        return openings.pipe(
            ops.group_join(
                source,
                closing_mapper,
                lambda _: empty(),
                ),
            ops.map(mapper),
            )
    return window_toggle


def _window(boundaries: Observable) -> Callable[[Observable], Observable]:
    """Projects each element of an observable sequence into zero or
    more windows.

    Args:
        source: Source observable to project into windows.

    Returns:
        An observable sequence of windows.
    """
    def window(source: Observable) -> Observable:

        def subscribe(observer, scheduler=None):
            window_subject = Subject()
            d = CompositeDisposable()
            r = RefCountDisposable(d)

            observer.on_next(add_ref(window_subject, r))

            def on_next_window(x):
                window_subject.on_next(x)

            def on_error(err):
                window_subject.on_error(err)
                observer.on_error(err)

            def on_completed():
                window_subject.on_completed()
                observer.on_completed()

            d.add(source.subscribe_(on_next_window, on_error, on_completed, scheduler))

            def on_next_observer(w):
                nonlocal window_subject
                window_subject.on_completed()
                window_subject = Subject()
                observer.on_next(add_ref(window_subject, r))

            d.add(boundaries.subscribe_(on_next_observer, on_error, on_completed, scheduler))
            return r
        return Observable(subscribe)
    return window


def _window_when(closing_mapper: Callable[[], Observable]) -> Callable[[Observable], Observable]:
    """Projects each element of an observable sequence into zero or
    more windows.

    Args:
        source: Source observable to project into windows.

    Returns:
        An observable sequence of windows.
    """
    def window_when(source: Observable) -> Observable:

        def subscribe(observer, scheduler=None):
            m = SerialDisposable()
            d = CompositeDisposable(m)
            r = RefCountDisposable(d)
            window = Subject()

            observer.on_next(add_ref(window, r))

            def on_next(value):
                window.on_next(value)

            def on_error(error):
                window.on_error(error)
                observer.on_error(error)

            def on_completed():
                window.on_completed()
                observer.on_completed()

            d.add(source.subscribe_(on_next, on_error, on_completed, scheduler))

            def create_window_on_completed():
                try:
                    window_close = closing_mapper()
                except Exception as exception:
                    observer.on_error(exception)
                    return

                def on_completed():
                    nonlocal window
                    window.on_completed()
                    window = Subject()
                    observer.on_next(add_ref(window, r))
                    create_window_on_completed()

                m1 = SingleAssignmentDisposable()
                m.disposable = m1
                m1.disposable = window_close.pipe(ops.take(1)).subscribe_(noop, on_error, on_completed, scheduler)

            create_window_on_completed()
            return r
        return Observable(subscribe)
    return window_when



from typing import Any, Callable, Optional

from rx import operators as ops
from rx.core import Observable
from rx.subject import BehaviorSubject
from rx.core.typing import Mapper


def _publish_value(initial_value: Any, mapper: Optional[Mapper] = None) -> Callable[[Observable], Observable]:
    if mapper:
        def subject_factory(scheduler):
            return BehaviorSubject(initial_value)

        return ops.multicast(subject_factory=subject_factory, mapper=mapper)
    return ops.multicast(BehaviorSubject(initial_value))



from abc import abstractmethod
from typing import Optional, List, Generator, Generic, TypeVar

import requests

from twitch.api import API
from twitch.baseresource import BaseResource

T = TypeVar('T')


class Resource(BaseResource, Generic[T]):

    def __init__(self, path: str, api: API, data: Optional[List[T]] = None, **kwargs):
        super().__init__(path=path, api=api, data=data, **kwargs)
        self._cursor: Optional[str] = None

    def __iter__(self) -> Generator[T, None, None]:
        # Yield available data
        if self._data:
            for entry in self._data:
                yield entry
            return

        # Check if iterator should paginate from api
        if self._can_paginate() is False:
            return

        # Remaining elements. Download all if set to None
        remaining: Optional[int] = self._kwargs.get('first', None)

        # When downloading add data, set first to maximum to minimize api calls
        if remaining is None:
            self._kwargs['first'] = BaseResource.FIRST_API_LIMIT

        # Paginate
        while remaining is None or remaining > 0:
            # Update remaining
            if remaining:
                self._kwargs['first'] = min(Resource.FIRST_API_LIMIT, remaining)

            # API Response
            try:
                response: dict = self._next_page()
            except requests.exceptions.HTTPError:
                return

            # Let resource handle pagination response and return elements
            elements: List[T] = self._handle_pagination_response(response)
            for element in elements:
                yield element

            # Decrement remaining if not None
            if remaining is not None:
                remaining = 0 if len(elements) > remaining else remaining - self._kwargs['first']

            # If no next cursor, stop
            if not self._cursor:
                break

    def __getitem__(self, index: int) -> T:
        if len(self._data) > index:
            return self._data[index]

        for i, value in enumerate(self._data):
            if i == index:
                return value

    @abstractmethod
    def _can_paginate(self) -> bool:
        """
        Check if resource can be paginated
        :return: Can paginate
        """
        return True

    @abstractmethod
    def _handle_pagination_response(self, response: dict) -> List[T]:
        """
        Pagination hook for every iteration
        :param response: Response from pagination
        :return: None
        """
        elements: List[T] = [T(api=self._api, data=data) for data in response['data']]

        return elements

    def _next_page(self, ignore_cache: bool = False) -> dict:
        """
        API Pagination
        Get next page from API
        :param ignore_cache: Whether to ignore API cache.
        :return: Response
        """
        # API Response
        self._kwargs['after'] = self._cursor

        response: dict = self._api.get(self._path, params=self._kwargs, ignore_cache=ignore_cache)

        # Set pagination cursor
        self._cursor = response.get('pagination', {}).get('cursor', None)

        return response



import sys
from typing import Callable, Optional

import rx
from rx.core import Observable
from rx.internal.utils import infinite


def _repeat(repeat_count: Optional[int] = None) -> Callable[[Observable], Observable]:
    if repeat_count is None:
        repeat_count = sys.maxsize

    def repeat(source: Observable) -> Observable:
        """Repeats the observable sequence a specified number of times.
        If the repeat count is not specified, the sequence repeats
        indefinitely.

        Examples:
            >>> repeated = source.repeat()
            >>> repeated = source.repeat(42)

        Args:
            source: The observable source to repeat.

        Returns:
            The observable sequence producing the elements of the given
            sequence repeatedly.
        """

        if repeat_count is None:
            gen = infinite()
        else:
            gen = range(repeat_count)

        return rx.defer(lambda _: rx.concat_with_iterable(source for _ in  gen))
    return repeat



from typing import List

from rx.disposable import Disposable, CompositeDisposable
from rx.core import Observable, typing
from rx.scheduler import VirtualTimeScheduler

from .subscription import Subscription



class ColdObservable(Observable):
    def __init__(self, scheduler: VirtualTimeScheduler, messages) -> None:
        super().__init__()

        self.scheduler: VirtualTimeScheduler = scheduler
        self.messages = messages
        self.subscriptions: List[Subscription] = []

    def _subscribe_core(self, observer=None, scheduler=None) -> typing.Disposable:
        self.subscriptions.append(Subscription(self.scheduler.clock))
        index = len(self.subscriptions) - 1
        disp = CompositeDisposable()

        def get_action(notification):
            def action(scheduler, state):
                notification.accept(observer)
                return Disposable()
            return action

        for message in self.messages:
            notification = message.value

            # Don't make closures within a loop
            action = get_action(notification)
            disp.add(self.scheduler.schedule_relative(message.time, action))

        def dispose() -> None:
            start = self.subscriptions[index].subscribe
            end = self.scheduler.to_seconds(self.scheduler.now)
            self.subscriptions[index] = Subscription(start, end)
            disp.dispose()

        return Disposable(dispose)



from typing import Callable

from rx.core import Observable
from rx.core.typing import Mapper
from rx.disposable import SerialDisposable, CompositeDisposable, SingleAssignmentDisposable
from rx.scheduler import ImmediateScheduler


def _expand(mapper: Mapper) -> Callable[[Observable], Observable]:
    def expand(source: Observable) -> Observable:
        """Expands an observable sequence by recursively invoking
        mapper.

        Args:
            source: Source obserable to expand.

        Returns:
            An observable sequence containing all the elements produced
            by the recursive expansion.
        """
        def subscribe(observer, scheduler=None):
            scheduler = scheduler or ImmediateScheduler.singleton()

            queue = []
            m = SerialDisposable()
            d = CompositeDisposable(m)
            active_count = [0]
            is_acquired = [False]

            def ensure_active():
                is_owner = False
                if queue:
                    is_owner = not is_acquired[0]
                    is_acquired[0] = True

                def action(scheduler, state):
                    if queue:
                        work = queue.pop(0)
                    else:
                        is_acquired[0] = False
                        return

                    sad = SingleAssignmentDisposable()
                    d.add(sad)

                    def on_next(value):
                        observer.on_next(value)
                        result = None
                        try:
                            result = mapper(value)
                        except Exception as ex:
                            observer.on_error(ex)
                            return

                        queue.append(result)
                        active_count[0] += 1
                        ensure_active()

                    def on_complete():
                        d.remove(sad)
                        active_count[0] -= 1
                        if active_count[0] == 0:
                            observer.on_completed()

                    sad.disposable = work.subscribe_(on_next, observer.on_error, on_complete, scheduler)
                    m.disposable = scheduler.schedule(action)

                if is_owner:
                    m.disposable = scheduler.schedule(action)

            queue.append(source)
            active_count[0] += 1
            ensure_active()
            return d
        return Observable(subscribe)
    return expand



from typing import List, Union, Generator, Tuple, Dict

import twitch.helix as helix
from twitch.api import API
from .resource import Resource


class Users(Resource['helix.User']):

    def __init__(self, api: API, *args):
        super().__init__(api=api, path='users')

        # Load data
        users: List[Union[str, int]] = []
        for user in args:
            users += [user] if type(user) in [str, int] else list(user)

        params: Dict[str, list] = dict()

        # Split user id and login (based on type t)
        params['id'], params['login'] = [
            list(set(n)) for n in [
                [str(user) for user in users if type(user) == t] for t in [int, str]
            ]
        ]

        # todo: Authenticated user if bearer token is provided
        if not len(params['id'] + params['login']):
            pass

        # Custom user caching
        if self._api.use_cache:
            cache_hits: Dict[str, list] = {'id': [], 'login': []}
            for key, users in tuple(params.items()):
                for user in users:
                    cache_key: str = f'helix.users.{key}.{user}'
                    cache_data: dict = API.SHARED_CACHE.get(cache_key)
                    if cache_data:
                        self._data.append(helix.User(api=self._api, data=cache_data))
                        cache_hits[key].append(user)

            # Remove cached users from params
            params['id'], params['login'] = [
                [n for n in params[key] if n not in cache_hits[key]] for key in ['id', 'login']
            ]

        # Fetch non-cached users from API
        if len(params['id'] + params['login']):
            for data in self._api.get(self._path, params=params, ignore_cache=True)['data']:

                # Create and append user)
                user = helix.User(api=self._api, data=data)
                self._data.append(user)

                # Save to cache
                if self._api.use_cache:
                    API.SHARED_CACHE.set(f'helix.users.login.{user.login}', data)
                    API.SHARED_CACHE.set(f'helix.users.id.{user.id}', data)


    def _can_paginate(self) -> bool:
        return False

    def _handle_pagination_response(self, response: dict) -> None:
        pass

    def _pagination_stream_done(self) -> None:
        pass

    def videos(self, **kwargs) -> Generator[Tuple['helix.User', 'helix.Videos'], None, None]:
        for user in self:
            yield user, user.videos(**kwargs)

    @property
    def streams(self) -> Generator[Tuple['helix.User', 'helix.Stream'], None, None]:
        for user in self:
            yield user, user.stream



from setuptools import setup, find_packages
 
classifiers = [
  'Development Status :: 3 - Alpha',
  'Intended Audience :: Other Audience',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]
 
setup(
  name="VODet",
  version = '0.0.2',
  description="A simple tool to gather emote analytics from any twitch VOD.",
  long_description=open('README.txt').read() + '\n\n' + open('CHANGELOG.txt').read(),
  url='',  
  author='Sabeeh Malik',
  author_email='sabeeh.malik@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='twitch', 
  packages=find_packages(),
  install_requires=['matplotlib','pyparsing','python-dateutil','twitch-python'] 
)


from typing import Callable
from rx.core import Observable
from rx.core.typing import Predicate


def _find_value(predicate: Predicate, yield_index) -> Callable[[Observable], Observable]:
    def find_value(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            i = [0]

            def on_next(x):
                should_run = False
                try:
                    should_run = predicate(x, i, source)
                except Exception as ex:  # pylint: disable=broad-except
                    observer.on_error(ex)
                    return

                if should_run:
                    observer.on_next(i[0] if yield_index else x)
                    observer.on_completed()
                else:
                    i[0] += 1

            def on_completed():
                observer.on_next(-1 if yield_index else None)
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return find_value



from typing import Any, Callable

import rx

from rx.core import Observable


def _start_with(*args: Any) -> Callable[[Observable], Observable]:
    def start_with(source: Observable) -> Observable:
        """Partially applied start_with operator.

        Prepends a sequence of values to an observable sequence.

        Example:
            >>> start_with(source)

        Returns:
            The source sequence prepended with the specified values.
        """
        start = rx.from_iterable(args)
        sequence = [start, source]
        return rx.concat(*sequence)
    return start_with



from typing import Callable
from rx.core import Observable


def _take_last(count: int) -> Callable[[Observable], Observable]:
    def take_last(source: Observable) -> Observable:
        """Returns a specified number of contiguous elements from the end of an
        observable sequence.

        Example:
            >>> res = take_last(source)

        This operator accumulates a buffer with a length enough to store
        elements count elements. Upon completion of the source sequence, this
        buffer is drained on the result sequence. This causes the elements to be
        delayed.

        Args:
            source: Number of elements to take from the end of the source
            sequence.

        Returns:
            An observable sequence containing the specified number of elements
            from the end of the source sequence.
        """

        def subscribe(observer, scheduler=None):
            q = []

            def on_next(x):
                q.append(x)
                if len(q) > count:
                    q.pop(0)

            def on_completed():
                while q:
                    observer.on_next(q.pop(0))
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return take_last



from typing import Callable

import rx
from rx.core import Observable



def _with_latest_from(*sources: Observable) -> Callable[[Observable], Observable]:
    """With latest from operator.

    Merges the specified observable sequences into one observable
    sequence by creating a tuple only when the first
    observable sequence produces an element. The observables can be
    passed either as seperate arguments or as a list.

    Examples:
        >>> op = with_latest_from(obs1)
        >>> op = with_latest_from(obs1, obs2, obs3)

    Returns:
        An observable sequence containing the result of combining
    elements of the sources into a tuple.
    """

    def with_latest_from(source: Observable) -> Observable:
        return rx.with_latest_from(source, *sources)
    return with_latest_from



from typing import Callable, Optional
from asyncio import Future

from .. import typing
from rx.core import Observable
from rx.internal.exceptions import SequenceContainsNoElementsError


def _to_future(future_ctor: Optional[Callable[[], Future]] = None,
               scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Future]:
    future_ctor = future_ctor or Future
    future = future_ctor()

    def to_future(source: Observable) -> Future:
        """Converts an existing observable sequence to a Future.

        If the observable emits a single item, then this item is set as the
        result of the future. If the observable emits a sequence of items, then
        the last emitted item is set as the result of the future.

        Example:
            future = rx.return_value(42).pipe(ops.to_future(asyncio.Future))

        Args:
            future_ctor: [Optional] The constructor of the future.

        Returns:
            A future with the last value from the observable sequence.
        """

        has_value = False
        last_value = None

        def on_next(value):
            nonlocal last_value
            nonlocal has_value
            last_value = value
            has_value = True

        def on_error(err):
            if not future.cancelled():
                future.set_exception(err)

        def on_completed():
            nonlocal last_value
            if not future.cancelled():
                if has_value:
                    future.set_result(last_value)
                else:
                    future.set_exception(SequenceContainsNoElementsError())
            last_value = None

        dis = source.subscribe_(on_next, on_error, on_completed, scheduler=scheduler)
        future.add_done_callback(lambda _: dis.dispose())

        return future
    return to_future



from typing import Callable

from rx.core import Observable


def _pairwise() -> Callable[[Observable], Observable]:
    def pairwise(source: Observable) -> Observable:
        """Partially applied pairwise operator.

        Returns a new observable that triggers on the second and
        subsequent triggerings of the input observable. The Nth
        triggering of the input observable passes the arguments from
        the N-1th and Nth triggering as a pair. The argument passed to
        the N-1th triggering is held in hidden internal state until the
        Nth triggering occurs.

        Returns:
            An observable that triggers on successive pairs of
            observations from the input observable as an array.
        """

        def subscribe(observer, scheduler=None):
            has_previous = [False]
            previous = [None]

            def on_next(x):
                pair = None

                with source.lock:
                    if has_previous[0]:
                        pair = (previous[0], x)
                    else:
                        has_previous[0] = True

                    previous[0] = x

                if pair:
                    observer.on_next(pair)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed)
        return Observable(subscribe)
    return pairwise



from typing import Callable, Any

from rx import operators as ops
from rx.core import Observable


def _do_while(condition: Callable[[Any], bool]) -> Callable[[Observable], Observable]:
    """Repeats source as long as condition holds emulating a do while
    loop.

    Args:
        condition: The condition which determines if the source will be
            repeated.

    Returns:
        An observable sequence which is repeated as long
        as the condition holds.
    """

    def do_while(source: Observable) -> Observable:
        return source.pipe(ops.concat(source.pipe(ops.while_do(condition))))
    return do_while



from typing import Callable, Optional

from rx.disposable import Disposable
from rx.core import typing
from rx.core import Observable
from rx.core.typing import Mapper


def _from_callback(func: Callable, mapper: Optional[Mapper] = None) -> Callable[[], Observable]:
    """Converts a callback function to an observable sequence.

    Args:
        func: Function with a callback as the last argument to
            convert to an Observable sequence.
        mapper: [Optional] A mapper which takes the arguments
            from the callback to produce a single item to yield on next.

    Returns:
        A function, when executed with the required arguments minus
        the callback, produces an Observable sequence with a single value of
        the arguments to the callback as a list.
    """

    def function(*args):
        arguments = list(args)

        def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> typing.Disposable:
            def handler(*args):
                results = list(args)
                if mapper:
                    try:
                        results = mapper(args)
                    except Exception as err: # pylint: disable=broad-except
                        observer.on_error(err)
                        return

                    observer.on_next(results)
                else:
                    if isinstance(results, list) and len(results) <= 1:
                        observer.on_next(*results)
                    else:
                        observer.on_next(results)

                    observer.on_completed()

            arguments.append(handler)
            func(*arguments)
            return Disposable()

        return Observable(subscribe)
    return function



from typing import Any, Callable, Iterable, Union, List

import rx
from rx.core import Observable, typing


def _combine_latest(*others: Observable) -> Callable[[Observable], Observable]:
    def combine_latest(source: Observable) -> Observable:
        """Merges the specified observable sequences into one
        observable sequence by creating a tuple whenever any
        of the observable sequences produces an element.

        Examples:
            >>> obs = combine_latest(source)

        Returns:
            An observable sequence containing the result of combining
            elements of the sources into a tuple.
        """

        sources = (source,) + others

        return rx.combine_latest(*sources)
    return combine_latest



from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Comparer
from rx.internal.basic import identity

from .min import first_only


def _max(comparer: Optional[Comparer] = None) -> Callable[[Observable], Observable]:
    """Returns the maximum value in an observable sequence according to
    the specified comparer.

    Examples:
        >>> op = max()
        >>> op = max(lambda x, y:  x.value - y.value)

    Args:
        comparer: [Optional] Comparer used to compare elements.

    Returns:
        An operator function that takes an observable source and returns
        an observable sequence containing a single element with the
        maximum element in the source sequence.
    """

    return pipe(
        ops.max_by(identity, comparer),
        ops.map(first_only)
    )



from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable
from rx.core.typing import Scheduler
from rx.scheduler import TimeoutScheduler
from rx.subject import AsyncSubject


def _to_async(func: Callable,
              scheduler: Optional[Scheduler] = None
              ) -> Callable:
    """Converts the function into an asynchronous function. Each
    invocation of the resulting asynchronous function causes an
    invocation of the original synchronous function on the specified
    scheduler.

    Examples:
        res = rx.to_async(lambda x, y: x + y)(4, 3)
        res = rx.to_async(lambda x, y: x + y, Scheduler.timeout)(4, 3)
        res = rx.to_async(lambda x: log.debug(x), Scheduler.timeout)('hello')

    Args:
        func: Function to convert to an asynchronous function.
        scheduler: [Optional] Scheduler to run the function on. If not
            specified, defaults to Scheduler.timeout.

    Returns:
        Aynchronous function.
    """

    _scheduler = scheduler or TimeoutScheduler.singleton()

    def wrapper(*args) -> Observable:
        subject = AsyncSubject()

        def action(scheduler, state):
            try:
                result = func(*args)
            except Exception as ex:  # pylint: disable=broad-except
                subject.on_error(ex)
                return

            subject.on_next(result)
            subject.on_completed()

        _scheduler.schedule(action)
        return subject.pipe(ops.as_observable())
    return wrapper



from typing import Callable, Optional, Sequence, Any

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Comparer
from rx.internal.basic import identity
from rx.internal.exceptions import SequenceContainsNoElementsError


def first_only(x: Sequence) -> Any:
    if not x:
        raise SequenceContainsNoElementsError()

    return x[0]


def _min(comparer: Optional[Comparer] = None) -> Callable[[Observable], Observable]:
    """The `min` operator.

    Returns the minimum element in an observable sequence according to
    the optional comparer else a default greater than less than check.

    Examples:
        >>> res = source.min()
        >>> res = source.min(lambda x, y: x.value - y.value)

    Args:
        comparer: [Optional] Comparer used to compare elements.

    Returns:
        An observable sequence containing a single element
        with the minimum element in the source sequence.
    """

    return pipe(
        ops.min_by(identity, comparer),
        ops.map(first_only)
    )



from typing import Any, Callable, Optional

from rx.core import typing
from rx.core import Observable
from rx.scheduler import CurrentThreadScheduler
from rx.core.abc.scheduler import Scheduler


def _return_value(value: Any, scheduler: Optional[typing.Scheduler] = None) -> Observable:
    """Returns an observable sequence that contains a single element,
    using the specified scheduler to send out observer messages.
    There is an alias called 'just'.

    Examples:
        >>> res = return(42)
        >>> res = return(42, rx.Scheduler.timeout)

    Args:
        value: Single element in the resulting observable sequence.

    Returns:
        An observable sequence containing the single specified
        element.
    """

    def subscribe(observer: typing.Observer, scheduler_: Optional[typing.Scheduler] = None) -> typing.Disposable:
        _scheduler = scheduler or scheduler_ or CurrentThreadScheduler.singleton()

        def action(scheduler: typing.Scheduler, state: Any = None):
            observer.on_next(value)
            observer.on_completed()

        return _scheduler.schedule(action)
    return Observable(subscribe)


def _from_callable(supplier: Callable[[], Any], scheduler: Optional[typing.Scheduler] = None) -> Observable:
    def subscribe(observer: typing.Observer, scheduler_: typing.Scheduler = None):
        _scheduler = scheduler or scheduler_ or CurrentThreadScheduler.singleton()

        def action(_: Scheduler, __: Any = None):
            nonlocal observer

            try:
                observer.on_next(supplier())
                observer.on_completed()
            except Exception as e:  # pylint: disable=broad-except
                observer.on_error(e)
        return _scheduler.schedule(action)

    return Observable(subscribe)



from typing import Callable, Optional
from rx.core import Observable
from rx.core.typing import Mapper, SubComparer
from rx.internal.basic import default_sub_comparer

from .minby import extrema_by


def _max_by(key_mapper: Mapper,
            comparer: Optional[SubComparer] = None
            ) -> Callable[[Observable], Observable]:

    cmp = comparer or default_sub_comparer

    def max_by(source: Observable) -> Observable:
        """Partially applied max_by operator.

        Returns the elements in an observable sequence with the maximum
        key value.

        Examples:
            >>> res = max_by(source)

        Args:
            source: The source observable sequence to.

        Returns:
            An observable sequence containing a list of zero or more
            elements that have a maximum key value.
        """
        return extrema_by(source, key_mapper, cmp)
    return max_by



from datetime import timedelta
from typing import List, Union

import twitch.v5 as v5
from twitch.api import API


class V5:
    BASE_URL: str = 'https://api.twitch.tv/v5/'

    def __init__(self, client_id: str,
                 client_secret: str = None,
                 request_rate: int = None,
                 use_cache: bool = False,
                 cache_duration: timedelta = timedelta(minutes=30)):
        self.client_secret: str = client_secret
        self.scope: List[str] = []

        self.api = API(V5.BASE_URL,
                       client_id,
                       use_cache=use_cache,
                       cache_duration=cache_duration,
                       request_rate=request_rate)

    def comments(self, video_id: Union[str, int]) -> 'v5.Comments':
        return v5.Comments(api=self.api, video_id=video_id)



import sys


class Subscription(object):
    def __init__(self, start, end=None):
        self.subscribe = start
        self.unsubscribe = end or sys.maxsize

    def equals(self, other):
        return self.subscribe == other.subscribe and self.unsubscribe == other.unsubscribe

    def __eq__(self, other):
        return self.equals(other)

    def __repr__(self):
        return str(self)

    def __str__(self):
        unsubscribe = "Infinite" if self.unsubscribe == sys.maxsize else self.unsubscribe
        return "(%s, %s)" % (self.subscribe, unsubscribe)



from abc import ABC, abstractmethod


class Observable(ABC):
    """Observable abstract base class. Untyped."""

    __slots__ = ()

    @abstractmethod
    def subscribe(self, observer=None, *, scheduler=None):
        raise NotImplementedError



from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Mapper


def _sum(key_mapper: Optional[Mapper] = None) -> Callable[[Observable], Observable]:
    if key_mapper:
        return pipe(
            ops.map(key_mapper),
            ops.sum()
        )

    return ops.reduce(seed=0, accumulator=lambda prev, curr: prev + curr)



from typing import Callable, Optional
from rx.core import Observable
from rx.core.typing import Mapper, Comparer
from rx.internal.basic import default_comparer


def array_index_of_comparer(array, item, comparer):
    for i, a in enumerate(array):
        if comparer(a, item):
            return i
    return -1


class HashSet:
    def __init__(self, comparer):
        self.comparer = comparer
        self.set = []

    def push(self, value):
        ret_value = array_index_of_comparer(self.set, value, self.comparer) == -1
        if ret_value:
            self.set.append(value)
        return ret_value


def _distinct(key_mapper: Optional[Mapper] = None,
              comparer: Optional[Comparer] = None
              ) -> Callable[[Observable], Observable]:
    comparer = comparer or default_comparer

    def distinct(source: Observable) -> Observable:
        """Returns an observable sequence that contains only distinct
        elements according to the key_mapper and the comparer. Usage of
        this operator should be considered carefully due to the maintenance
        of an internal lookup structure which can grow large.

        Examples:
            >>> res = obs = distinct(source)

        Args:
            source: Source observable to return distinct items from.

        Returns:
            An observable sequence only containing the distinct
            elements, based on a computed key value, from the source
            sequence.
        """

        def subscribe(observer, scheduler=None):
            hashset = HashSet(comparer)

            def on_next(x):
                key = x

                if key_mapper:
                    try:
                        key = key_mapper(x)
                    except Exception as ex:
                        observer.on_error(ex)
                        return

                hashset.push(key) and observer.on_next(x)
            return source.subscribe_(on_next, observer.on_error,observer.on_completed, scheduler)
        return Observable(subscribe)
    return distinct



from typing import TypeVar, Generic, List, Optional

from .api import API

T = TypeVar('T')


class BaseResource(Generic[T]):
    FIRST_API_LIMIT: int = 100

    def __init__(self, path: str, api: API, data: Optional[List[T]] = None, **kwargs):
        self._path: str = path
        self._api: Optional[API] = api
        self._data: List[T] = data or []
        self._kwargs: dict = kwargs

    def __iter__(self):
        for item in self._data:
            yield item

    def __getitem__(self, index: int) -> T:
        return self._data[index]



from concurrent.futures import Future, ThreadPoolExecutor
from typing import Optional

from rx.core.abc import Startable
from rx.core import typing

from .newthreadscheduler import NewThreadScheduler


class ThreadPoolScheduler(NewThreadScheduler):
    """A scheduler that schedules work via the thread pool."""

    class ThreadPoolThread(Startable):
        """Wraps a concurrent future as a thread."""

        def __init__(self,
                     executor: ThreadPoolExecutor,
                     target: typing.StartableTarget):
            self.executor: ThreadPoolExecutor = executor
            self.target: typing.StartableTarget = target
            self.future: Optional[Future] = None

        def start(self) -> None:
            self.future = self.executor.submit(self.target)

        def cancel(self) -> None:
            if self.future:
                self.future.cancel()

    def __init__(self, max_workers: Optional[int] = None) -> None:
        self.executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=max_workers)

        def thread_factory(target: typing.StartableTarget):
            return self.ThreadPoolThread(self.executor, target)

        super().__init__(thread_factory)

